//Webベースの自由対話システム
// 引用元 : 柴田雅博, 冨浦洋一, 西口友美, "雑談自由対話を実現するためのＷＷＷ上の文書からの妥当な候補文選択手法", 人工知能学会論文誌 24(6), 507-519, 2009.
// https://www.jstage.jst.go.jp/article/tjsai/24/6/24_6_507/_pdf
//
// 作成 : 宮澤幸希（m-kouki@moegi.waseda.jp）
//   ※ 本プログラムは　柴田, 冨浦 & 西口, 2009.　の引用及び解釈に基づきます。
//
//【引数】
//  DialogSystem(ジャンル名, ユーザの発話の認識結果);
//
//【実行例】
//(初回)
// java HariboteDialogSystem(アーティスト);
//(2回目以降)
// java HariboteDialogSystem(アーティスト　椎名林檎は元気ですか？);
//
// ※初回は「(ジャンル名)の話をしましょう」と出力します。
//   このとき、「ユーザの発話の認識結果」は指定しないでください（引数が1個なら初回と判定します）
// ※システムの発話文は、このプログラムのあるフォルダ内の「outputSystemSpeak.txt」に書き出されます。

import java.io.*;
import java.util.*;

public class DialogSystem {
	private static String mainThemeFilePath = "";			//メインテーマを列挙したテキストファイルのパス
	private static String systemSpeakFileName = "";			//システムの出力文
	private static String logFileName = "./log.txt";		//ログ
	private static String usedFileName = "./used.txt";		//使用済み文リスト
	private static String centerFileName = "./center.txt";	//前回の中心語
	private static String themeFileName = "./theme.txt";	//この会話のメインテーマ
	private static String corpusDirName = "./corpus";		//コーパスファイルが入ったフォルダのパス
	private static String fileType = "UTF-8";
	
	public static void main(String[] args) {
    	String genreName = "";	//ジャンル名
    	String userSpeak = "";	//ユーザの発話の認識結果
    	String center = "";		//現在の中心語
    	String theme = "";		//今回の会話のテーマ   
    	int dialogNum = 1;
    	if (args.length == 3) {
    		//初回の会話（システムが喋るだけ）
    		genreName = args[0];
    		mainThemeFilePath = args[1];
    		systemSpeakFileName = args[2];
    		// 初回と判断して会話数をリセット
    		dialogNum = setDialogNum(true);
    	} else if (args.length == 4) {
    		//二回目以降の会話（ユーザ発話に対する応答）
    		genreName = args[0];
    		mainThemeFilePath = args[1];
    		systemSpeakFileName = args[2];
    		userSpeak = args[3];
    		// この会話が初回から何回目なのかを取得
    		dialogNum = setDialogNum(false);
        	// ユーザー発話文を出力
        	System.out.println("あなた : " + userSpeak);
        	// 現在の中心語を取得
        	center = getCenter(userSpeak);
        	// メインテーマを取得
        	theme = getTheme();
    	} else {

    	}
    	//システム発話文を生成
    	String systemSpeak = makeSystemSpeak(genreName, userSpeak, dialogNum, center, theme);
    	System.out.println("システム : " + systemSpeak + " (" + dialogNum + ")(" + center + ")(" + theme + ")");
    	//システム発話文を出力
    	outputFile(systemSpeak, systemSpeakFileName, fileType, false);
//【未実装】システムが「へー」を返した場合、「へー」前後のユーザ発話をまとめて一つの発話と判断して、次の返答文を計算する
//【未実装】「意味的整合度の合計が-2.5にならなかった場合、候補文を使わず『へー』を返す」は未実装です（整合度の計算にlogを使っていないので）    				
    	//ログを出力
    	outputFile(dialogNum + "\t" + systemSpeak + "\t" + center, logFileName, fileType, true);
    	//使用済み文リストを出力
    	outputFile(systemSpeak, usedFileName, fileType, true);
    	//前回の中心語を出力
    	outputFile(center, centerFileName, fileType, false);
    }
    
    //システム発話文を生成
    public static String makeSystemSpeak(String genreName, String userSpeak, int dialogNum, String center, String theme) {
    	SearchCorpus sc = new SearchCorpus();
    	String systemSpeak = "";
    	try {
    		if (dialogNum == 1) {
    			systemSpeak = genreName + "の話をしましょう。";
    		} else {
    			//メインテーマが決まっていなければユーザ初回発話と判断する
    			if (theme.equals("")) {
        			//STEP1 : ユーザの初回発話の中心語がメインテーマリストにある語かどうかをチェック
        			boolean centerCheck = checkMainTheme(center);
        			if (centerCheck == false) {
        				//なかったとき、リテイク
        				systemSpeak = genreName + "の中で、興味のあるものはなんですか？";
        			} else {
        				//あったとき
        				System.out.println( "この対話のメインテーマ : " + center + "　に決定!");
        				theme = center;
        		    	// メインテーマをファイルに出力!
        				outputFile(center, themeFileName, fileType, true);
        				// システム発話を作成する
        				systemSpeak = sc.selectSystemSpeech(userSpeak, corpusDirName, usedFileName, fileType, center, theme, genreName);
        			}
    			} else {
    				//システム発話を作成する
    				systemSpeak = sc.selectSystemSpeech(userSpeak, corpusDirName, usedFileName, fileType, center, theme, genreName);
    			}
    			//systemSpeak = "へーそうですか。";	
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return systemSpeak;
    }
    
    //現在の中心語を取得
    public static String getCenter(String userSpeak){
    	String center = "";
    	try {
    		//ユーザ発話文の中心語を解析する
    		center = searchCenterWord(userSpeak);
    		//中心語が見つからなかったら、前回の中心語を使う
    		if (center.equals("")) {
        	    BufferedReader brk = new BufferedReader( new InputStreamReader( new FileInputStream(centerFileName) , fileType) );
        	    String line = "";
        	    while((line = brk.readLine()) != null) {
        	    	center = line;	//center.txtの1行目を取得している、もっとうまいコードの書き方があるかも
        	    	break;
        	    }
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return center;
    }
    
    //メインテーマを取得
    public static String getTheme(){
    	String theme = "";
    	try {
    	    BufferedReader brk = new BufferedReader( new InputStreamReader( new FileInputStream(themeFileName) , fileType) );
    	    String line = "";
    	    while((line = brk.readLine()) != null) {
    	    	theme = line;	//theme.txtの1行目を取得している、もっとうまいコードの書き方があるかも
    	    	break;
    	    }
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return theme;
    }
    
    //指定した文の中心語を調べる
    // センタリング理論：ある文の話題の中心となる語を選ぶ規則
    // 1. ゼロ代名詞のときの、前の文の「～が」「～は」の名詞 (このメソッドは""を返します)
    // 2. 「～は」の名詞
    // 3. 「～が」の名詞
    // 4. 「～を」の名詞
    public static String searchCenterWord(String line){
    	String centerWord = "";
    	try {
    		//形態素解析
//【タスク】事前にSenの辞書に固有名詞（特にメインテーマ）を追加しておく！
    		// SENを実行

    		String senresult = SenSample.runSen(line);
    		//System.out.println(senresult);
    		String spl1[] = senresult.split("\n");
    		String beforeWord = "";
    		String beforeHinshi = "";
    		boolean gaSwitch = false;
    		for (int count1 = 0 ; count1 < spl1.length ; count1++) {
    			String spl2[] = spl1[count1].split("\\(");
    			String spl3[] = spl2[1].split(",");
    			//System.out.println("語 : " + spl2[0] + "\t 品詞 : " + spl3[0]);
    			//名詞＋助詞のセットを探す
    			if( ( spl3[0].equals("助詞")) && ( beforeHinshi.equals("名詞") ) && ( spl2[0].equals("は") ) ) {
    				centerWord = beforeWord;
    				break;
    			} else if( ( spl3[0].equals("助詞")) && ( beforeHinshi.equals("名詞") ) && ( spl2[0].equals("が") ) ) {
    				centerWord = beforeWord;
    				gaSwitch = true;
    			} else if( ( spl3[0].equals("助詞")) && ( beforeHinshi.equals("名詞") ) && ( spl2[0].equals("を") ) ) {
    				if (gaSwitch != true) {
    					centerWord = beforeWord;
    				}
    			}
    			beforeWord = spl2[0];
    			beforeHinshi = spl3[0];
    		}
   		
    		//centerWord = "テスト用中心語です。";
    		
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return centerWord;
    }
    
    //対話回数を判定（初回だったら各種設定ファイルを初期化する）
    public static int setDialogNum(boolean firstCheck) {
    	int dialogNum = 1;
    	if (firstCheck == false ){
    		//2回目以降のとき
    		dialogNum = 2;
    	} else {
    		//初回のとき
    		outputFile("", logFileName, fileType, false);		//ログを初期化
    		outputFile("", usedFileName, fileType, false);		//使用済み文リストを初期化
    		outputFile("", centerFileName, fileType, false);	//前回の中心語を初期化
    		outputFile("", themeFileName, fileType, false);		//メインテーマを初期化
    	}
    	return dialogNum;
    }
    
    //ファイル書き出し（引数：書き出すテキスト、ファイル名、ファイルの文字コード、新規か追記か（新規ならfalse））
    public static void outputFile(String text, String fileName, String fileType, boolean overWriteCheck) {
    	try {
    		if (overWriteCheck == false) {
    			//新規書き出し
                OutputStreamWriter out = new OutputStreamWriter( new FileOutputStream( fileName ), fileType );
                if(text.equals("")) {
                    out.write( text );	//何も書き出さないときは改行もしない
                } else {
                    out.write( text + "\n" );
                }
                out.close();
    		} else {
    			//追記
                OutputStreamWriter out = new OutputStreamWriter( new FileOutputStream( fileName, true ), fileType );
                out.write( text + "\n" );
                out.close();
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    }
    
    //ユーザの初回発話の中心語がメインテーマリストにある語かどうかをチェック
    public static boolean checkMainTheme(String center){
    	boolean result = false;
    	try {
    		//メインテーマを列挙したテキストファイルを読み込む
    	    BufferedReader brk = new BufferedReader( new InputStreamReader( new FileInputStream(mainThemeFilePath) , fileType) );
    	    String line = "";
    	    while((line = brk.readLine()) != null) {
    	    	if (line.equals(center)) {
    	    		result = true;
    	    		break;
    	    	}
    	    }
    	    brk.close();
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
		return result;
    }
}
