//Webベースの自由対話システム - DialogSystem.javaの入出力を再現したプログラム
// システムの応答は「へーそうなんだ」のみ
//
// 引用元 : 柴田雅博, 冨浦洋一, 西口友美, "雑談自由対話を実現するためのＷＷＷ上の文書からの妥当な候補文選択手法", 人工知能学会論文誌 24(6), 507-519, 2009.
// https://www.jstage.jst.go.jp/article/tjsai/24/6/24_6_507/_pdf
//
// 作成 : 宮澤幸希（m-kouki@moegi.waseda.jp）
//   ※ 本プログラムは　柴田, 冨浦 & 西口, 2009.　の引用及び解釈に基づきます。
//
//【引数】
//  HariboteDialogSystem(ジャンル名, ユーザの発話の認識結果);
//
//【実行例】
//(初回)
// java HariboteDialogSystem(アーティスト);
//(2回目以降)
// java HariboteDialogSystem(アーティスト　椎名林檎は元気ですか？);
//
// ※初回は「(ジャンル名)の話をしましょう」と出力します。
//   このとき、「ユーザの発話の認識結果」は指定しないでください（引数が1個なら初回と判定します）
// ※システムの発話文は、このプログラムのあるフォルダ内の「outputSystemSpeak.txt」に書き出されます。

import java.io.*;

public class HariboteDialogSystem {
    public static void main(String[] args) {
    	String genreName = "";	//ジャンル名
    	String userSpeak = "";	//ユーザの発話の認識結果
    	int dialogNum = 1;
    	if (args.length == 1) {
    		genreName = args[0];
    		//初回と判断して会話数をリセット
    		dialogNum = getDialogNum(true);
    	} else {
    		genreName = args[0];
    		userSpeak = args[1];
    		//この会話が初回から何回目なのかを取得
    		dialogNum = getDialogNum(false);
    	}
    	System.out.println("あなた : " + userSpeak);
    	//システム発話文を生成
    	String systemSpeak = makeSystemSpeak(genreName, userSpeak, dialogNum);
    	System.out.println("システム : " + systemSpeak + " (" + dialogNum + ")");
    	//システム発話文を出力
    	outputSystemSpeak(systemSpeak, "UTF-8");
    }
    //対話回数を判定
    public static int getDialogNum(boolean firstCheck) {
    	int dialogNum = 1;
    	if (firstCheck == false ){
    		dialogNum = 2;
    	}
    	return dialogNum;
    }
    //システム発話文を生成
    public static String makeSystemSpeak(String genreName, String userSpeak, int dialogNum) {
    	String systemSpeak = "";
    	try {
    		if (dialogNum == 1) {
    			systemSpeak = genreName + "の話をしましょう。";
    		} else {
    			systemSpeak = "へーそうですか。";
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return systemSpeak;
    }
    //システム発話文を出力
    public static void outputSystemSpeak(String outputSystemSpeak, String fileType) {
    	try {
            OutputStreamWriter out = new OutputStreamWriter( new FileOutputStream( "./outputSystemSpeak.txt" ), fileType );
            out.write( outputSystemSpeak + "\n" );
            out.close();
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    }
}
