//Webベースの自由対話システム - Webコーパス作成
// 引用元 : 柴田雅博, 冨浦洋一, 西口友美, "雑談自由対話を実現するためのＷＷＷ上の文書からの妥当な候補文選択手法", 人工知能学会論文誌 24(6), 507-519, 2009.
// https://www.jstage.jst.go.jp/article/tjsai/24/6/24_6_507/_pdf
//
// 作成 : 宮澤幸希（m-kouki@moegi.waseda.jp）
//   ※ 本プログラムは　柴田, 冨浦 & 西口, 2009.　の引用及び解釈に基づきます。
//
//候補文コーパス作成
// 1. コーパスのジャンルを決める（例：「映画」）
// 2. メインテーマを決める（例：「ゴッドファーザー」）
// 3. YahooAPIを使ってWebページを収集する（クエリ：「映画 ゴッドファーザー」）
//    このとき、HTMLのタイトルにメインテーマを含むもののみを対象にする
// 4. 各文を切り出して、メインテーマ、文書ID（URL）、文ID（その文書の何番目の文か）を付加して書き出す。
//    このとき、各文の中心語も書き出しておきます。また、内容語の数を数えて、3以上のものだけ出力します。
//
//【コーパスファイルの仕様】 [本文]\t[メインテーマ]\t[文書ID(URL)]\t[文ID]\t[中心語]\t[カンマ区切りの内容語一覧]
//【注意】事前にメインテーマを列挙したテキストファイルを作っておきます。

import java.io.*;
import java.net.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

public class MakeCorpus {
	static String genre = "ゾイド";									//ジャンル
	//static String genre = "アーティスト";								//ジャンル
	static String corpusDirName = "./corpus";						//コーパスファイルを保存するフォルダのパス
	static String mainThemeFilePath = "./mainTheme_zoids.txt";		//メインテーマを列挙したテキストファイルのパス
	//static String mainThemeFilePath = "./mainTheme_artist.txt";	//メインテーマを列挙したテキストファイルのパス
	static String fileType = "UTF-8";								//ファイルの文字コード 
	static int getResultsNum = 300;									//検索結果を何件まで取得するか
	
    public static void main(String[] args) {
    	if (args.length == 5) {
    		genre = args[0];
    		corpusDirName = args[1];
    		mainThemeFilePath = args[2];
    		fileType = args[3];
    		getResultsNum = Integer.parseInt(args[4]);
    	} else {
    		System.out.println("引数の数が足りません。引数は [ジャンル名] [コーパスファイルを格納するディレクトリ名] [メインテーマを列挙したテキストファイルのパス] [ファイルの文字コード] [検索結果を何件まで取得するか] の順に指定してください。");
    		System.out.println("今回はデフォルトパラメータ " + genre + " でコーパスを作成します..."); 
    	}
    	try {
    		//メインテーマを列挙したテキストファイルを読み込む
    		// 以下では、テキストファイルの文字コードに「SJIS」を指定して読み込んでいます。文字コードは自由に書き直してください。
		    BufferedReader brk = new BufferedReader( new InputStreamReader( new FileInputStream(mainThemeFilePath) , fileType) );
		    String mainTheme = "";
		    while((mainTheme = brk.readLine()) != null) {
		    	System.out.println("ジャンル : " + genre + "\t メインテーマ : " + mainTheme);
		    	webCrawl(mainTheme);
		    }
		    brk.close();
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    }
    public static void webCrawl(String mainTheme) {
		//コーパス出力用のディレクトリを生成
		File mydir1 = new File(corpusDirName);
		mydir1.mkdir();
		
    	String query = genre + " " + mainTheme;
    	String appid = "*******";  				// YahooAPI で登録した自分のアプリケーションIDを入力（2014年時点で無効）
    	try {
			//書き出すコーパスファイルのパス
			String corpusFilePath = corpusDirName + "/" + genre + "_" + mainTheme + ".txt" ;
    		//ファイル初期化
            OutputStreamWriter out0 = new OutputStreamWriter( new FileOutputStream(corpusFilePath) , fileType);
            out0.write("");
            out0.close();

    		query = URLEncoder.encode(query, "UTF-8");	// クエリの文字コードをUTF-8に変換する
    		//検索WebサービスのURL（アプリケーションIDとクエリを指定して接続する）
    		URL url = new URL("http://search.yahooapis.jp/WebSearchService/V1/webSearch?appid="
    				+ appid + "&query=" + query+ "&results=" + getResultsNum);
    		System.out.println(url);
    		URLConnection uc = url.openConnection();
    		// 結果を読み出すためのストリーム
    		InputStream is = uc.getInputStream();
    		// 検索結果のタイトルだけ取得
    		String[] result = getAssocWords(is, getResultsNum);
    		for (int count = 0 ; count < result.length ; count++) {
	        	String thisResult = result[count];
	        	String spl[] = thisResult.split("\t");
	        	//HTMLのタイトルにメインテーマを含むとき
	        	if (spl[0].indexOf(mainTheme) != -1){
	        		String thisTitle = spl[0];
	        		String thisUrl = spl[1];
	        		System.out.println("タイトル : " + thisTitle + "\t URL : " + thisUrl);
	        		//本文を取得する
	        		ParseHTML html = new ParseHTML();
	        		String article = html.parseHTML(thisUrl);
// 課題 ： エラーが出て解析できないページがある。
	        		//System.out.println("\t" + article);
	        		//各文を切り出し
	        		String splBrake1[] = article.split("\n");
	        		int totalLineCount = 0;
	        		String beforeCenterWord = "NOTHING";
	        		//さまざまな改行記号で分割
	        		for (int countB1 = 0 ; countB1 < splBrake1.length ; countB1++) {
		        		String splBrake2[] = splBrake1[countB1].split("\r\n");
		        		for (int countB2 = 0 ; countB2 < splBrake2.length ; countB2++) {
			        		String splBrake3[] = splBrake2[countB2].split("\r");
			        		for (int countB3 = 0 ; countB3 < splBrake3.length ; countB3++) {
				        		String splBrake4[] = splBrake3[countB3].split("。");
				        		for (int countB4 = 0 ; countB4 < splBrake4.length ; countB4++) {
					        		String splBrake5[] = splBrake4[countB4].split("．");
					        		for (int countB5 = 0 ; countB5 < splBrake5.length ; countB5++) {
						        		String splBrake6[] = splBrake5[countB5].split("？");
						        		for (int countB6 = 0 ; countB6 < splBrake6.length ; countB6++) {
							        		String splBrake7[] = splBrake6[countB6].split("\\?");
							        		for (int countB7 = 0 ; countB7 < splBrake7.length ; countB7++) {
								        		String splBrake8[] = splBrake7[countB7].split("！");
								        		for (int countB8 = 0 ; countB8 < splBrake8.length ; countB8++) {
									        		String splBrake9[] = splBrake8[countB8].split("\\!");
									        		for (int countB9 = 0 ; countB9 < splBrake9.length ; countB9++) {
									        			if(splBrake9[countB9].equals("")) {
									        			} else {
									        				//切り出されたテキスト
									        				String articleLine = splBrake9[countB9];
//		課題 ： articleLine の文字化け対処!!
									        				//String articleLine = URLEncoder.encode(splBrake9[countB9], "UTF-8");
									        				
									        				//切り出されたテキストの中心語を判定
									        				String centerWord = searchCorpusCenterWord(articleLine, beforeCenterWord);
									        				//文頭の中心語が存在しないテキスト、または文でないと判定されたテキストはコーパスに保存しない
//		【独自処理】本文中にコロンや括弧など、読めないテキストがあるものは削除する
									        				if ( centerWord.equals("NOTSENTENCE") || centerWord.equals("NOTHING") ||
									        						( articleLine.indexOf(":") != -1) || ( articleLine.indexOf("：") != -1) || 
								        							( articleLine.indexOf("@") != -1) || ( articleLine.indexOf("【") != -1) ||
								        							( articleLine.indexOf("[") != -1) || ( articleLine.indexOf("/") != -1)) {
									        				} else {
									        					//内容語の数を数えて、3以上のものだけ出力する
									        					int contentWordNum = countContentWord(articleLine);
									        					if ( contentWordNum >= 3) {
											        				System.out.println("\t" + articleLine);
											        				beforeCenterWord = centerWord;
											        				//メインテーマ、文書ID（URL）、文ID（その文書の何番目の文か）、中心語、カンマ区切りの内容語一覧 を付加して書き出す
											        				String contentWordList = listContentWord(articleLine);
											        				
											        				//同じ候補文がすでにコーパスに登録されていないかをチェックして、新規なら書き出す
											        				boolean checkDuplication = checkDuplicationCorpus(articleLine, corpusFilePath);
											        				if ( checkDuplication == false ){
												        				OutputStreamWriter out = new OutputStreamWriter( new FileOutputStream( corpusFilePath, true) , fileType);
												        				out.write(articleLine + "\t" + mainTheme + "\t" + thisUrl + "\t" + totalLineCount + "\t" + centerWord + "\t" + contentWordList + "\n");
												        	            out.close();
											        				}
											        				totalLineCount++;
									        					}
									        				}
									        			}
									        		}
								        		}
							        		}
						        		}
					        		}
				        		}
			        		}
		        		}
	        		}
	        	}
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    }
    //YahooAPIで取得した検索結果のXMLを解析する
    // XMLの仕様は以下のとおり。
    //  <ResultSet (略)> 
    //    <Result> 
    //    <Title>タイトル</Title> 
    //    <Summary>スニペット</Summary> 
    //    <Url>http://www.oricon.co.jp/prof/artist/30824/</Url> 
    //    <ClickUrl>(略)</ClickUrl> 
    //    <ModificationDate>1251730800</ModificationDate> 
    //    <MimeType>text/html</MimeType> 
    //    <Cache> 
    //      <Url>(略)</Url> 
    //      <Size>61947</Size> 
    //    </Cache> 
    //    </Result> 
    //    <Result> 
    //    (略)
    //    </Result>
    //  </ResultSet> 
    //  <!-- pls723.search.kcd.yahoo.co.jp compressed/chunked Fri Nov 13 11:43:55 JST 2009 --> 
    private static String[] getAssocWords(InputStream is, int getResultsNum) {
    	//タイトル \t URL の形式で出力する
        String[] pageList = new String[getResultsNum];
        try {    
        	DocumentBuilderFactory dbfactory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = dbfactory.newDocumentBuilder();
            Document doc = builder.parse(is);
            // ルートノード
            Element root = doc.getDocumentElement();
            // Titleノードを収集
            NodeList listTitle = root.getElementsByTagName("Title");
            for (int count = 0; count < listTitle.getLength(); count++) {
                Element elm = (Element) listTitle.item(count);
                pageList[count] = elm.getFirstChild().getNodeValue();
            }
            // ClickUrlノードを収集
            NodeList listUrl = root.getElementsByTagName("ClickUrl");
            for (int count = 0; count < listUrl.getLength(); count++) {
                Element elm = (Element) listUrl.item(count);
                pageList[count] = pageList[count] + "\t" + elm.getFirstChild().getNodeValue();
            }
        } catch (Exception e) {
        	System.out.println(e);
        }
        return pageList;
    }
    //指定した文の中心語を調べる
    // センタリング理論：ある文の話題の中心となる語を選ぶ規則
    // 1. ゼロ代名詞のときの、前の文の「～が」「～は」の名詞
    // 2. 「～は」の名詞
    // 3. 「～が」の名詞
    // 4. 「～を」の名詞
    public static String searchCorpusCenterWord(String line, String beforeCenterWord) {
    	String centerWord = beforeCenterWord;
    	try {
    		//形態素解析
 //タスク：辞書に固有名詞（特にメインテーマ）を追加！
    		String senresult = SenSample.runSen(line);
    		//System.out.println(senresult);
    		String spl1[] = senresult.split("\n");
    		String beforeWord = "";
    		String beforeHinshi = "";
    		boolean gaSwitch = false;
    		int countContentWord = 0;
 //【独自処理】品詞が3つに満たないものは文とみなさない（ソースコードの断片のようなテキストが多くあるため）
    		if (spl1.length < 3) {
    			centerWord = "NOTSENTENCE";
    		} else {
        		for (int count1 = 0 ; count1 < spl1.length ; count1++) {
        			String spl2[] = spl1[count1].split("\\(");
        			String spl3[] = spl2[1].split(",");
        			//System.out.println("語 : " + spl2[0] + "\t 品詞 : " + spl3[0]);
        			//内容語の数を数える
        			if ( spl3[0].equals("名詞") || spl3[0].equals("動詞") || spl3[0].equals("形容詞") || spl3[0].equals("形容動詞") ) {
        				countContentWord++;
        			}
            		//名詞＋助詞のセットを探す
        			if( ( spl3[0].equals("助詞")) && ( beforeHinshi.equals("名詞") ) && ( spl2[0].equals("は") ) ) {
        				//確定！
        				centerWord = beforeWord;
        				break;
        			} else if( ( spl3[0].equals("助詞")) && ( beforeHinshi.equals("名詞") ) && ( spl2[0].equals("が") ) ) {
        				centerWord = beforeWord;
        				gaSwitch = true;
        			} else if( ( spl3[0].equals("助詞")) && ( beforeHinshi.equals("名詞") ) && ( spl2[0].equals("を") ) ) {
        				if (gaSwitch != true) {
        					centerWord = beforeWord;
        				}
        			}
            		beforeWord = spl2[0];
            		beforeHinshi = spl3[0];
        		}
//【独自処理】内容語がひとつもないものは文とみなさない
        		if (countContentWord == 0) {
        			centerWord = "NOTSENTENCE";
        		}
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return centerWord;
    }
    //指定した文の内容語の数を数える
    public static int countContentWord(String texts) {
    	int result = 0;
    	try {
       		String senresult = SenSample.runSen(texts);
    		String spl1[] = senresult.split("\n");
    		for (int count1 = 0 ; count1 < spl1.length ; count1++) {
    			String spl2[] = spl1[count1].split("\\(");
    			String spl3[] = spl2[1].split(",");
    			//System.out.println("語 : " + spl2[0] + "\t 品詞 : " + spl3[0]);
        		//名詞＋助詞のセットを探す
    			if ( spl3[0].equals("名詞") || spl3[0].equals("動詞") || spl3[0].equals("形容詞") || spl3[0].equals("形容動詞") ) {
    				result++;
    			}
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return result;
    }
    //指定した文の内容語の一覧（カンマ区切り）を出力する
    public static String listContentWord(String texts) {
    	String result = "";
    	try {
       		String senresult = SenSample.runSen(texts);
    		String spl1[] = senresult.split("\n");
    		for (int count1 = 0 ; count1 < spl1.length ; count1++) {
    			String spl2[] = spl1[count1].split("\\(");
    			String spl3[] = spl2[1].split(",");
    			//System.out.println("語 : " + spl2[0] + "\t 品詞 : " + spl3[0]);
        		//名詞＋助詞のセットを探す
    			if ( spl3[0].equals("名詞") || spl3[0].equals("動詞") || spl3[0].equals("形容詞") || spl3[0].equals("形容動詞") ) {
    				result = result + spl2[0] + ",";
    			}
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return result;
    }
    
	//指定した候補文がすでにコーパスに登録されていないかをチェック
    public static boolean checkDuplicationCorpus(String articleLine, String corpusFilePath){
    	boolean result = false;
    	try {
		    BufferedReader brk = new BufferedReader( new InputStreamReader( new FileInputStream(corpusFilePath) , fileType) );
		    String line = "";
		    while((line = brk.readLine()) != null) {
		    	String spl[] = line.split("\t");
		    	if(spl[0].equals(articleLine)) {
		    		result = true;
		    	}
		    }
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return result;
    }
}
