//Webベースの自由対話システム - HTMLパーサ
// システム全体の引用元 : 柴田雅博, 冨浦洋一, 西口友美, "雑談自由対話を実現するためのＷＷＷ上の文書からの妥当な候補文選択手法", 人工知能学会論文誌 24(6), 507-519, 2009.
// https://www.jstage.jst.go.jp/article/tjsai/24/6/24_6_507/_pdf
//
// 本プログラム（HTMLパーサ）の引用元 : SimpleParser3
//   http://www.mlab.im.dendai.ac.jp/~yamada/ir/HTMLParser/SimpleParser3.html
//
// 作成 : 宮澤幸希（m-kouki@moegi.waseda.jp）
//   ※ 本プログラムは　柴田, 冨浦 & 西口, 2009.　および SimpleParser3 の引用及び解釈に基づきます。

import org.htmlparser.*;
import org.htmlparser.util.*;
import org.htmlparser.tags.*;
import org.htmlparser.nodes.*;

public class ParseHTML {
    public static String parseHTML (String url) {
    	String outputLine = "";
        Parser parser = new Parser ();
        NodeFilter filter = null;
    	try {
			//URLを解析する
    		parser.setResource (url);
    		NodeList list = parser.parse(filter);
    		NodeIterator i = list.elements ();
            while (i.hasMoreNodes ()) {
            	String textLine = "";
            	textLine = textLine + processMyNodes(i.nextNode());
            	//タグを削除する
            	textLine = removeTags(textLine);
                //特定の記号を削除する（ブログ作者の癖にあわせる必要があるかも）
                textLine = removeStrings(textLine, "\t");
                textLine = removeStrings(textLine, "&nbsp;");
                textLine = removeStrings(textLine, " ");
            	//System.out.println(textLine + "\n");
            	outputLine = outputLine + textLine;
    	    }
    	} catch (Exception err) {
			System.out.println(err);
		}
    	return outputLine;
    }
    static String processMyNodes (Node node) throws ParserException {
    	String textLine = "";
    	try {
            if (node instanceof TextNode) {
                TextNode text = (TextNode)node;		// Node から TextNode にキャスト
                textLine = text.getText();			//本文のテキストを取得
            } else if (node instanceof TagNode) {
                TagNode tag = (TagNode)node;
                if(tag instanceof LinkTag) {
                	LinkTag link = (LinkTag)tag;	// 相対URLは絶対URLに変換される
                	//System.out.println("Link: " + link.getLink());
                }
                NodeList nl = tag.getChildren ();
                if (null != nl) {
                    for (NodeIterator i = nl.elements (); i.hasMoreNodes(); ) {
                    	textLine = textLine + processMyNodes (i.nextNode());
                    }
                }
            }
    	} catch (Exception err) {
			System.out.println(err);
		}
        return textLine;
    }
    //指定された文字列を削除する
	public static String removeStrings(String line, String deleteStrings) {
		String outputLine = line;
		try {
			int position = 0;
			int counter = 0;
			while(position > -1) {
				position = line.indexOf(deleteStrings, position);
				if (position > -1) {	// deleteStrings が存在するとき
					StringBuffer sb = new StringBuffer(line);
					sb.delete(position, (position + deleteStrings.length()) );
					line = sb.toString();
				}
				//エラー対策強制終了
				if (counter > 10000) {
					break;
				}
				counter++;
			}
			outputLine = line;
		} catch (Exception err) {
			System.out.println(err);
		}
		return outputLine;
	}
	//タグ（<から>まで）を取り除く
	public static String removeTags(String line) {
		String outputLine = line;
		try {
			//タグを全て取り除く
			int position = 0;
			int counter = 0;
			while(position > -1) {
				//開始タグの位置を調べる
				position = line.indexOf("<", position);
				//System.out.println("< が " + position + " 文字目に出現!");
				if (position > -1) {	// < が存在するとき
					// > の位置を調べる
					int endpos = line.indexOf(">", (position + 1));
					//System.out.println("> が " + endpos + " 文字目に出現!");
					//タグの部分を削除
					StringBuffer sb = new StringBuffer(line);
					sb.delete(position, (endpos + 1) );
					line = sb.toString();
				}
				//エラー対策強制終了
				if (counter > 10000) {
					break;
				}
				counter++;
			}
			outputLine = line;
		} catch (Exception err) {
			System.out.println(err);
		}
		return outputLine;
	}
}
