//Webベースの自由対話システム - MakeCorpusで作ったコーパスファイルを解析して、ユーザ発話の応答文を選択
// システム全体の引用元 : 柴田雅博, 冨浦洋一, 西口友美, "雑談自由対話を実現するためのＷＷＷ上の文書からの妥当な候補文選択手法", 人工知能学会論文誌 24(6), 507-519, 2009.
// https://www.jstage.jst.go.jp/article/tjsai/24/6/24_6_507/_pdf
//
// 作成 : 宮澤幸希（m-kouki@moegi.waseda.jp）
//   ※ 本プログラムは　柴田, 冨浦 & 西口, 2009.　および SimpleParser3 の引用及び解釈に基づきます。

import java.io.*;
import java.util.*;

public class SearchCorpus {
	//実行テスト用
	public static void main(String[] args) {
		//selectSystemSpeech("ガンスナイパーが好きです", "./corpus", "./used.txt", "UTF-8", "ガンスナイパー", "ガンスナイパー", "ゾイド");
		selectSystemSpeech("ジェノブレイカーが好きです", "./corpus", "./used.txt", "UTF-8", "ジェノブレイカー", "ジェノブレイカー", "ゾイド");
		//selectSystemSpeech("椎名林檎は元気ですか？", "./corpus", "./used.txt", "UTF-8", "椎名林檎", "椎名林檎", "アーティスト");
	}
	
	//DialogSystemから呼び出されるメソッド
	public static String selectSystemSpeech(String userSpeak, String corpusDirName, String usedFileName, String fileType, String center, String theme, String genreName) {
		String output = "";
		try {
			//コーパスファイルのパスを取得
			String corpusFilePath = corpusDirName + "/" + genreName + "_" + theme + ".txt" ;
			//ユーザ発話の内容語を取得する
			String userContentList = listContentWord(userSpeak);
			//今回のユーザ発話の中心語と同じ中心語のテキストを選び、 意味的整合度を求めて、cache.txtに書き出す
			searchCenterCorpusText(center, userContentList, corpusFilePath, fileType);
			//候補文の中から、一つを決める
			output = selectSystemText("./cache.txt", center, usedFileName, fileType);
			
			//【未実装】文書間の意味的整合度が最も高く、かつその値が -2.5 以上の文が、候補文に選ばれる。条件を満たす文がなかった場合、メインテーマを中心語にして候補文選択をやりなおす。
			//それでも条件を満たす文がなかった場合、システム発話は「へー」となる。
			if (output.equals("")) {
				output = "へー";
			}
			//【未実装】選ばれた文をゼロ代名詞化する。
			//【未実装】選ばれた文の冒頭に接続詞があった場合、「ちなみに」「それにしても」「そういえば」「例えば」「そりゃ」「本当は」を除いて除去する
			
			System.out.println( "システム発話文 : " + output );
		} catch (Exception e) {
    	    System.out.println(e);
    	}
		return output;
	}
	
    //コーパスの中から、ユーザ発話と同じ中心語の文を検索して、意味的整合度を求めて、 cache.txt に書き出す
	// cache.txt の形式 : [候補文no.]\t[候補文]\t[各内容語とユーザ発話との意味的整合度のリスト（カンマ区切り）]
    public static void searchCenterCorpusText(String center, String userContentList, String corpusFilePath, String fileType){
    	try {
    		OutputStreamWriter out0 = new OutputStreamWriter( new FileOutputStream( "cache.txt") , fileType);
    		out0.write("");
    		out0.close();
    		
    		OutputStreamWriter out = new OutputStreamWriter( new FileOutputStream( "cache.txt", true) , fileType);

    	    BufferedReader brk = new BufferedReader( new InputStreamReader( new FileInputStream(corpusFilePath) , fileType) );
    	    String line = "";
    	    int countText = 1;
    	    while((line = brk.readLine()) != null) {
    	    	String spl[] = line.split("\t");
    	    	if ( spl[4].equals(center) ) {
    	    		System.out.println( "候補文 no." + countText + " : " + spl[0] + " (内容語 ... " + spl[5] + ")");
    	    		out.write(countText + "\t" + spl[0] + "\t");
    	    		//候補文の内容語を順番に処理する
       	    		String splSysContList[] = spl[5].split(",");
    	    		for (int countCont = 0 ; countCont < splSysContList.length ; countCont++) {
    	    			String systemContent = splSysContList[countCont];
        				//候補文の内容語が、中心語でないときのみ処理する
        				if (systemContent.equals(center)) {
        				} else {
            				//候補文の内容語とユーザ発話との意味的整合度を求める
            				double match = calcMatching(userContentList, systemContent, center, corpusFilePath, fileType);
            				System.out.println( "\t" + systemContent + "とユーザ発話との意味的整合度 ： " + match );
            				out.write(match + ",");
        				}
    	    		}
    	    		out.write("\n");
    	    		countText++;
    	    	}
    	    }
    	    brk.close();
            
            out.close();
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    }
    
    //指定した語とユーザ発話との意味的整合度を求める
    // ユーザ発話の内容語（名詞、動詞、形容詞、形容動詞）ωと、システム発話の内容語ω'の意味的整合度は、
    // r(ω,ω') = logP(ω'|ω) - logP_D(ω') で決まる。
    // 		P(ω'|ω) は、ωを含む文があらわれたとき、その次の文がω'を含む確率
    //      P(ω'|ω) は最尤推定で求める。具体的には、連続する2文を抜き出していき 「文1がωを含み、文2がω'を含んでいる数」÷「文1がωを含んでいる数」 で求める。
    //		P_D(ω') は、文書集合Dの中で、ω'を含む割合 df(ω')/|D| Dはウェブページ単位
    //		なお、ωを含む文の数が4以下なら、r(ω,ω')= -10000 として扱う。
 //【独自処理】「『ユーザ発話の（中心語を除いた）各内容語』と『候補文の特定の内容語』の組み合わせの中で、最も値が大きかった意味的整合度一つが出力されます。
    public static double calcMatching(String userContentList, String systemContent, String center, String corpusFilePath, String fileType) {
    	double result = -10000;
    	try {
    		//ユーザ発話の内容語を順番に処理する
    		String splUsrContList[] = userContentList.split(",");
    		//候補文の内容語を順番に処理する
    		for (int countCont = 0 ; countCont < splUsrContList.length ; countCont++) {
    			String userContent = splUsrContList[countCont];
				//ユーザ発話の内容語が、中心語でないときのみ、内容語同士の意味的整合度を求める
				if (userContent.equals(center)) {
				} else {
    				// 1. [ユーザ発話の内容語] を含む文が出現したとき、その次の文が [候補文の内容語] を含む確率
					double countUall = 0;			// [ユーザ発話の内容語] を含む文が出現する回数
    				double countMatchUafterS = 0;	// 上記のうち次の文が [候補文の内容語] を含む回数
    				boolean uAfterCheck = false;
    				// 2. 各ウェブページが [候補文の内容語] を含む割合		
    	    	    String beforePageUrl = "";
    	    	    double countPageNum = 0;
    	    	    double mutchPageNum = 0;
    	    	    boolean thisPageMutchCheck = false;		//同じページ内で該当する内容語がすでに出現したかどうか

    	    	    BufferedReader brk = new BufferedReader( new InputStreamReader( new FileInputStream(corpusFilePath) , fileType) );
    	    	    String line = "";
    	    	    while((line = brk.readLine()) != null) {
    	    	    	String spl[] = line.split("\t");
    	    	    	//【独自処理】中心語が一致したテキストのみを対象にする
    	    	    	if (spl[1].equals(center)) {
        	    	    	if (uAfterCheck == true) {
            	    	    	if ( spl[5].indexOf(systemContent) != -1 ) {
            	    	    		//System.out.println("\t\t[ユーザ発話の内容語] を含む文が出現したとき、その次の文が [候補文の内容語] を含む例を発見!");
            	    	    		countMatchUafterS = countMatchUafterS + 1;
            	    	    	}
        	    	    		uAfterCheck = false;
        	    	    	}
        	    	    	if ( spl[0].indexOf(userContent) != -1 ) {
        	    	    		uAfterCheck = true;
        	    	    		countUall = countUall + 1;
        	    	    	}
        	    	    	
        	    	    	if (beforePageUrl.equals(spl[2])) {
        	    	    	} else {
        	    	    		//新しいページが出現
        	    	    		countPageNum = countPageNum + 1;
        	    	    		thisPageMutchCheck = false;
        	    	    	}
        	    	    	if ( ( spl[0].indexOf(systemContent) != -1 ) && (thisPageMutchCheck == false) ) {
        	    	    		mutchPageNum = mutchPageNum + 1;
        	    	    		thisPageMutchCheck = true;
        	    	    	}
        	    	    	beforePageUrl = spl[2];
    	    	    	}
    	    	    }
    	    	    brk.close();

    	    	    double logP = 0;
    	    	    if(countUall == 0) {	// [ユーザ発話の内容語] を含む文が出現しなかったとき
    	    	    	logP = 0;
    	    	    } else {
    	    	    	logP = Math.log(countMatchUafterS/countUall);	// logP(ω'|ω)
    	    	    }

    	    	    double logPd = 0;
    	    	    if(countPageNum == 0) {	// ページ数がゼロのとき
    	    	    	logPd = 0;
    	    	    } else {
    	    	    	logPd = Math.log( mutchPageNum / countPageNum );
    	    	    }
    	    	    //System.out.println(mutchPageNum);
    	    	    //System.out.println(countPageNum);
    	    	    //System.out.println(( mutchPageNum / countPageNum ));
    	    	    //System.out.println(Math.log( mutchPageNum / countPageNum ));
    	    	    
    	    	    double seigou = 0;
    	    	    if(mutchPageNum < 5) {	// ωを含む文の数が4以下のとき
    	    	    	seigou = -10000;
    	    	    } else {
    	    	    	seigou = logP - logPd;
    	    	    }
    	    	    //最も値の大きい意味的整合度を出力する
    	    	    if (result < seigou) {
    	    	    	result = seigou;
    	    	    }
    	    	    System.out.println("\t\tユーザ発話の内容語 : " + userContent + "\t 候補文の内容語 : " + systemContent + "\t 意味的整合度 ： " + result );
    				
				}
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return result;
    }
    
    //システムの候補文リストの中から、一つを決める
    public static String selectSystemText(String listFileName, String center, String usedfilePath, String fileType){
    	String output = "";
    	double maxMatching = -10000.0;		//最大の意味的整合度
    	try {
    		//候補文と意味的整合度のリストを読み込む
    	    BufferedReader brk = new BufferedReader( new InputStreamReader( new FileInputStream(listFileName) , fileType) );
    	    String line = "";
    	    while((line = brk.readLine()) != null) {
    	    	String spl[] = line.split("\t");
    	    	String seigouList = spl[2];
	    		//意味的整合度の上位3つを合計する
	    		double totalSeigou = 0;
	    		//System.out.println( "\t" + seigouList );
	    		String splSeigou[] = seigouList.split(",");
	    		if (splSeigou.length < 3) {
	    			totalSeigou = -10000;
	    		} else {
		    		double[] splSeigouD = new double[splSeigou.length];
		    		for (int countSeigou = 0 ; countSeigou < splSeigou.length ; countSeigou++) {
		    			splSeigouD[countSeigou] = Double.parseDouble(splSeigou[countSeigou]);
		    		}
		    		Arrays.sort(splSeigouD);	//配列をソート
		    		//for (int count = 0 ; count < splSeigouD.length ; count++) {
		    		//	System.out.println(splSeigouD[count]);
		    		//}
		    		totalSeigou = splSeigouD[splSeigou.length-1] + splSeigouD[splSeigou.length-2] + splSeigouD[splSeigou.length-3];
	    		}
				//System.out.println( "\t意味的整合度の合計 ： " + totalSeigou );
	    		
				if (maxMatching < totalSeigou) {
					//使用済みリストにすでに書かれているかどうかをチェック
		    	    boolean usedCheck = false;
		    	    
		    	    BufferedReader brkUsed = new BufferedReader( new InputStreamReader( new FileInputStream(usedfilePath) , fileType) );
		    	    String lineUsed = "";
		    	    while((lineUsed = brkUsed.readLine()) != null) {
		    	    	if (lineUsed.equals(spl[1])) {
		    	    		usedCheck = true;
		    	    	}
		    	    }
		    	    brkUsed.close();
		    	    
		    	    if (usedCheck == false) {
		    	    	output = spl[1];
		    	    	maxMatching = totalSeigou;
		    	    }
				}
    	    }
    	    brk.close();
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return output;
    }
    
    //指定した文の内容語の一覧（カンマ区切り）を出力する
    public static String listContentWord(String texts) {
    	String result = "";
    	try {
       		String senresult = SenSample.runSen(texts);
    		String spl1[] = senresult.split("\n");
    		for (int count1 = 0 ; count1 < spl1.length ; count1++) {
    			String spl2[] = spl1[count1].split("\\(");
    			String spl3[] = spl2[1].split(",");
    			//System.out.println("語 : " + spl2[0] + "\t 品詞 : " + spl3[0]);
        		//名詞＋助詞のセットを探す
    			if ( spl3[0].equals("名詞") || spl3[0].equals("動詞") || spl3[0].equals("形容詞") || spl3[0].equals("形容動詞") ) {
    				result = result + spl2[0] + ",";
    			}
    		}
    	} catch (Exception e) {
    	    System.out.println(e);
    	}
    	return result;
    }
}
